import logging
import bugsnag


class BugsnagHandler(logging.Handler):
    def emit(self, record):
        print record.__dict__.keys()
        if record.exc_info:
            bugsnag.notify(record.exc_info)



import inspect
import traceback

import bugsnag


MAX_STRING_LENGTH = 1024


def sanitize_object(obj, **kwargs):
    filters = kwargs.get("filters", [])

    if isinstance(obj, dict):
        clean_dict = {}
        for k, v in obj.iteritems():
            # Remove values for keys matching filters
            if any(f in k for f in filters):
                clean_dict[k] = "[FILTERED]"
            else:
                clean_obj = sanitize_object(v, **kwargs)
                if clean_obj:
                    clean_dict[k] = clean_obj

        return clean_dict
    elif any(isinstance(obj, t) for t in (list, set, tuple)):
        return [sanitize_object(x, **kwargs) for x in obj]
    else:
        try:
            if isinstance(obj, unicode):
                string = obj
            else:
                string = unicode(str(obj), errors='replace')

        except Exception:
            exc = traceback.format_exc()
            bugsnag.warn("Could not add object to metadata: %s" % exc)
            string = "[BADENCODING]"

        return string[:MAX_STRING_LENGTH]


def fully_qualified_class_name(obj):
    module = inspect.getmodule(obj)
    if module is not None and module.__name__ != "__main__":
        return module.__name__ + "." + obj.__class__.__name__
    else:
        return obj.__class__.__name__


def package_version(package_name):
    try:
        import pkg_resources
    except ImportError:
        return None
    else:
        try:
            return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            return None



from flask import got_request_exception, request, session

import bugsnag
from bugsnag.wsgi import request_path


def handle_exceptions(app):
    got_request_exception.connect(__log_exception, app)


# pylint: disable-msg=W0613
def __log_exception(sender, exception, **extra):
    bugsnag.configure_request(
        context="%s %s" % (request.method, request_path(request.environ)),
        user_id=request.remote_addr,
        request_data={
            "url": request.base_url,
            "headers": dict(request.headers),
            "cookies": dict(request.cookies),
            "params": dict(request.form),
        },
        session_data=dict(session),
        environment_data=dict(request.environ),
    )

    bugsnag.auto_notify(exception)

    bugsnag.clear_request_config()



from distutils.sysconfig import get_python_lib
import threading

from bugsnag.utils import fully_qualified_class_name


threadlocal = threading.local()


class _BaseConfiguration(object):
    def get(self, name, overrides=None):
        """
        Get a single configuration option, using values from overrides
        first if they exist.
        """
        if overrides:
            return overrides.get(name, getattr(self, name))
        else:
            return getattr(self, name)

    def configure(self, **options):
        """
        Set one or more configuration settings.
        """
        for name, value in options.items():
            if hasattr(self, name):
                setattr(self, name, value)

        return self


class Configuration(_BaseConfiguration):
    """
    Global app-level Bugsnag configuration settings.
    """
    def __init__(self):
        self.api_key = None
        self.release_stage = "production"
        self.notify_release_stages = None
        self.auto_notify = True
        self.use_ssl = False
        self.lib_root = get_python_lib()
        self.project_root = None
        self.app_version = None
        self.params_filters = ["password", "password_confirmation"]
        self.ignore_classes = []
        self.endpoint = "notify.bugsnag.com"

    def should_notify(self):
        return self.notify_release_stages is None or \
            self.release_stage in self.notify_release_stages

    def should_ignore(self, exception):
        return self.ignore_classes is not None and \
            fully_qualified_class_name(exception) in self.ignore_classes

    def get_endpoint(self):
        proto = "https" if self.use_ssl else "http"
        return "%s://%s" % (proto, self.endpoint)


class RequestConfiguration(_BaseConfiguration):
    """
    Per-request Bugsnag configuration settings.
    """
    @classmethod
    def get_instance(cls):
        """
        Get this thread's instance of the RequestConfiguration.
        """
        instance = getattr(threadlocal, "bugsnag", None)
        if not instance:
            instance = RequestConfiguration()
            setattr(threadlocal, "bugsnag", instance)

        return instance

    @classmethod
    def clear(cls):
        """
        Clear this thread's instance of the RequestConfiguration.
        """
        if hasattr(threadlocal, "bugsnag"):
            delattr(threadlocal, "bugsnag")

    def __init__(self):
        self.context = None
        self.user_id = None
        self.extra_data = {}
        self.request_data = {}
        self.environment_data = {}
        self.session_data = {}



import sys

from bugsnag.configuration import Configuration, RequestConfiguration
from bugsnag.notification import Notification


configuration = Configuration()


def configure(**options):
    """
    Configure the Bugsnag notifier application-wide settings.
    """
    configuration.configure(**options)


def configure_request(**options):
    """
    Configure the Bugsnag notifier per-request settings.
    """
    RequestConfiguration.get_instance().configure(**options)


def clear_request_config():
    """
    Clears the per-request settings.
    """
    RequestConfiguration.clear()


def notify(exception, **options):
    """
    Notify bugsnag of an exception.
    """
    if isinstance(exception, (list, tuple)):
        # Exception tuples, eg. from sys.exc_info
        if not "traceback" in options:
            options["traceback"] = exception[2]

        Notification(exception[1], configuration,
                     RequestConfiguration.get_instance(), **options).deliver()
    else:
        # Exception objects
        Notification(exception, configuration,
                     RequestConfiguration.get_instance(), **options).deliver()


def auto_notify(exception, **options):
    """
    Notify bugsnag of an exception if auto_notify is enabled.
    """
    if configuration.auto_notify:
        notify(exception, **options)


def log(message):
    """
    Print a log message with a Bugsnag prefix.
    """
    print("** [Bugsnag] %s" % message)


def warn(message):
    """
    Print a warning message with a Bugsnag prefix.
    """
    sys.stderr.write("** [Bugsnag] WARNING: %s\n" % message)


# Hook into all uncaught exceptions
def __bugsnag_excepthook(exctype, exception, traceback):
    auto_notify(exception, traceback=traceback)
    _old_excepthook(exctype, exception, traceback)

_old_excepthook = sys.excepthook
sys.excepthook = __bugsnag_excepthook



from werkzeug.wrappers import Request

import bugsnag
from bugsnag.wsgi import request_path


def handle_exception(exception, env):
    request = Request(env)

    bugsnag.configure_request(
        context="%s %s" % (request.method, request_path(env)),
        user_id=request.remote_addr,
        request_data={
            "url": request.base_url,
            "headers": dict(request.headers),
            "cookies": dict(request.cookies),
            "params": dict(request.form),
        },
        environment_data=dict(request.environ),
    )
    bugsnag.auto_notify(exception)
    bugsnag.clear_request_config()


class BugsnagMiddleware(object):
    """
    Bugsnag middleware bitches
    """

    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        try:
            iterable = self.application(environ, start_response)
        except Exception as e:
            handle_exception(e, environ)
            raise

        try:
            for event in iterable:
                yield event
        except Exception as e:
            handle_exception(e, environ)
            raise
        finally:
            if iterable and hasattr(iterable, 'close') and \
               callable(iterable.close):

                try:
                    iterable.close()
                except Exception:
                    handle_exception(e, environ)



from django.conf import settings

import bugsnag


def get_user_id(request):
    return request.META['REMOTE_ADDR']


def is_development_server(request):
    server = request.META.get('wsgi.file_wrapper', None)

    if server is None:
        return False

    return server.__module__ == 'django.core.servers.basehttp'


class BugsnagMiddleware(object):
    def __init__(self):
        # Ignore django 404s by default
        bugsnag.configuration.ignore_classes.append("django.http.Http404")

        # Import Bugsnag settings from settings.py
        django_bugsnag_settings = getattr(settings, 'BUGSNAG', {})
        bugsnag.configure(**django_bugsnag_settings)

    # pylint: disable-msg=R0201
    def process_request(self, request):
        if is_development_server(request):
            bugsnag.configure(release_stage="development")

        try:
            bugsnag.configure_request(
                context=request.path,
                user_id=get_user_id(request),
                session_data=dict(request.session),
                request_data={
                    'path': request.path,
                    'encoding': request.encoding,
                    'params': dict(request.REQUEST),
                    'url': request.build_absolute_uri(),
                },
                environment_data=dict(request.META),
            )

        except Exception as exc:
            bugsnag.log("Error in request middleware: %s" % exc)

        return None

    # pylint: disable-msg=W0613
    def process_response(self, request, response):
        bugsnag.clear_request_config()

        return response

    def process_exception(self, request, exception):
        try:
            bugsnag.auto_notify(exception)

        except Exception as exc:
            bugsnag.log("Error in exception middleware: %s" % exc)

        bugsnag.clear_request_config()

        return None



from tornado.web import RequestHandler
import bugsnag


class BugsnagRequestHandler(RequestHandler):
    def _handle_request_exception(self, exc):
        # Set the request info
        bugsnag.configure_request(
            user_id=self.request.remote_ip,
            context=self._get_context(),
            request_data={
                "url": self.request.full_url(),
                "method": self.request.method,
                "arguments": self.request.arguments,
            },
        )

        # Notify bugsnag
        bugsnag.notify(exc)

        # Call the parent handler
        RequestHandler._handle_request_exception(self, exc)

    def _get_context(self):
        return "%s %s" % (self.request.method, self.request.uri.split('?')[0])



import os
import urllib2
import sys
import threading
import traceback

try:
    import json
except ImportError:
    import simplejson as json

import bugsnag
from bugsnag.utils import sanitize_object
from bugsnag.utils import fully_qualified_class_name as class_name
from bugsnag.utils import package_version


def request(req):
    try:
        resp = urllib2.urlopen(req)
        status = resp.getcode()

        if status != 200:
            bugsnag.log("Notification to %s failed, status %d" % status)

    except Exception:
        bugsnag.log("Notification to %s failed" % (req.get_full_url()))
        print traceback.format_exc()


class Notification(object):
    """
    A single exception notification to Bugsnag.
    """
    NOTIFIER_NAME = "Python Bugsnag Notifier"
    NOTIFIER_URL = "https://github.com/bugsnag/bugsnag-python"

    def __init__(self, exception, config, request_config, **options):
        self.exception = exception
        self.options = options
        self.config = config
        self.request_config = request_config

    def deliver(self):
        """
        Deliver the exception notification to Bugsnag.
        """
        url = self.config.get_endpoint()

        try:
            if self.config.api_key is None:
                bugsnag.log("No API key configured, couldn't notify")
                return

            # Return early if we shouldn't notify for current release stage
            if not self.config.should_notify():
                return

            # Return early if we should ignore exceptions of this type
            if self.config.should_ignore(self.exception):
                return

            # Generate the payload and make the request
            bugsnag.log("Notifying %s of exception" % url)

            payload = self.__generate_payload()
            req = urllib2.Request(url, payload, {
                'Content-Type': 'application/json'
            })
            threading.Thread(target=request, args=(req,)).start()

        except Exception:
            exc = traceback.format_exc()
            bugsnag.warn("Notification to %s failed:\n%s" % (url, exc))

    def __generate_payload(self):
        try:
            # Set up the lib root
            lib_root = self.config.get("lib_root", self.options)
            if lib_root and lib_root[-1] != os.sep:
                lib_root += os.sep

            # Set up the project root
            project_root = self.config.get("project_root", self.options)
            if project_root and project_root[-1] != os.sep:
                project_root += os.sep

            # Build the stacktrace
            tb = self.options.get("traceback", sys.exc_info()[2])
            if tb:
                trace = traceback.extract_tb(tb)
            else:
                trace = traceback.extract_stack()

            bugsnag_module_path = os.path.dirname(bugsnag.__file__)

            stacktrace = []
            for line in trace:
                file_name = os.path.abspath(str(line[0]))
                in_project = False

                if file_name.startswith(bugsnag_module_path):
                    continue

                if lib_root and file_name.startswith(lib_root):
                    file_name = file_name[len(lib_root):]
                elif project_root and file_name.startswith(project_root):
                    file_name = file_name[len(project_root):]
                    in_project = True

                stacktrace.append({
                    "file": file_name,
                    "lineNumber": int(str(line[1])),
                    "method": str(line[2]),
                    "inProject": in_project,
                })

            stacktrace.reverse()

            # Fetch the notifier version from the package
            notifier_version = package_version("bugsnag_python") or "unknown"

            # Construct the payload dictionary
            payload = {
                "apiKey": self.config.api_key,
                "notifier": {
                    "name": self.NOTIFIER_NAME,
                    "url": self.NOTIFIER_URL,
                    "version": notifier_version,
                },
                "events": [{
                    "releaseStage": self.config.get("release_stage", self.options),
                    "appVersion": self.config.get("app_version", self.options),
                    "context": self.request_config.get("context", self.options),
                    "userId": self.request_config.get("user_id", self.options),
                    "exceptions": [{
                        "errorClass": class_name(self.exception),
                        "message": str(self.exception),
                        "stacktrace": stacktrace,
                    }],
                    "metaData": self.__generate_metadata(),
                }]
            }

            # JSON-encode and return the payload
            return json.dumps(payload)
        finally:
            del tb

    def __generate_metadata(self):
        return {
            "request": sanitize_object(
                self.request_config.get("request_data", self.options)
            ),
            "environment": sanitize_object(
                self.request_config.get("environment_data", self.options)
            ),
            "session": sanitize_object(
                self.request_config.get("session_data", self.options)
            ),
            "extraData": sanitize_object(
                self.request_config.get("extra_data", self.options)
            ),
        }
